@@ -267,6 +267,8 @@ class CalibratedLatticeEnsembleConfig(_Config, _HasFeatureConfigs,
                num_lattices=None,
                lattice_rank=None,
                interpolation='hypercube',
+               parameterization='all_vertices',
+               num_terms=2,
                separate_calibrators=True,
                use_linear_combination=False,
                use_bias=False,
@@ -276,6 +278,7 @@ class CalibratedLatticeEnsembleConfig(_Config, _HasFeatureConfigs,
                output_calibration=False,
                output_calibration_num_keypoints=10,
                output_initialization='quantiles',
+               output_calibration_input_keypoints_type='fixed',
                fix_ensemble_for_2d_constraints=True,
                random_seed=0):
     # pyformat: disable
@@ -305,6 +308,34 @@ class CalibratedLatticeEnsembleConfig(_Config, _HasFeatureConfigs,
         'simplex' uses d+1 parameters and thus scales better. For details see
         `tfl.lattice_lib.evaluate_with_simplex_interpolation` and
         `tfl.lattice_lib.evaluate_with_hypercube_interpolation`.
+      parameterization: The parameterization of the lattice function class to
+        use. A lattice function is uniquely determined by specifying its value
+        on every lattice vertex. A parameterization scheme is a mapping from a
+        vector of parameters to a multidimensional array of lattice vertex
+        values. It can be one of:
+          - String `'all_vertices'`: This is the "traditional" parameterization
+            that keeps one scalar parameter per lattice vertex where the mapping
+            is essentially the identity map. With this scheme, the number of
+            parameters scales exponentially with the number of inputs to the
+            lattice. The underlying lattices used will be `tfl.layers.Lattice`
+            layers.
+          - String `'kronecker_factored'`: With this parameterization, for each
+            lattice input i we keep a collection of `num_terms` vectors each
+            having `feature_configs[0].lattice_size` entries (note that all
+            features must have the same lattice size). To obtain the tensor of
+            lattice vertex values, for `t=1,2,...,num_terms` we compute the
+            outer product of the `t'th` vector in each collection, multiply by a
+            per-term scale, and sum the resulting tensors. Finally, we add a
+            single shared bias parameter to each entry in the sum. With this
+            scheme, the number of parameters grows linearly with `lattice_rank`
+            (assuming lattice sizes and `num_terms` are held constant).
+            Currently, only monotonicity shape constraint and bound constraint
+            are supported for this scheme. Regularization is not currently
+            supported. The underlying lattices used will be
+            `tfl.layers.KroneckerFactoredLattice` layers.
+      num_terms: The number of terms in a lattice using `'kronecker_factored'`
+        parameterization. Ignored if parameterization is set to
+        `'all_vertices'`.
       separate_calibrators: If features should be separately calibrated for each
         lattice in the ensemble.
       use_linear_combination: If set to true, a linear combination layer will be
@@ -330,6 +361,11 @@ class CalibratedLatticeEnsembleConfig(_Config, _HasFeatureConfigs,
           - String `'uniform'`: Output is initliazed uniformly in label range.
           - A list of numbers: To be used for initialization of the output
             lattice or output calibrator.
+      output_calibration_input_keypoints_type: One of "fixed" or
+        "learned_interior". If "learned_interior", keypoints are initialized to
+        the values in `pwl_calibration_input_keypoints` but then allowed to vary
+        during training, with the exception of the first and last keypoint
+        location which are fixed.
       fix_ensemble_for_2d_constraints: A boolean indicating whether to add
         missing features to some lattices to resolve potential 2d constraint
         violations which require lattices from ensemble to either contain both
@@ -375,12 +411,16 @@ class CalibratedLatticeConfig(_Config, _HasFeatureConfigs,
   def __init__(self,
                feature_configs=None,
                interpolation='hypercube',
+               parameterization='all_vertices',
+               num_terms=2,
                regularizer_configs=None,
                output_min=None,
                output_max=None,
                output_calibration=False,
                output_calibration_num_keypoints=10,
-               output_initialization='quantiles'):
+               output_initialization='quantiles',
+               output_calibration_input_keypoints_type='fixed',
+               random_seed=0):
     """Initializes a `CalibratedLatticeConfig` instance.
 
     Args:
@@ -392,6 +432,34 @@ class CalibratedLatticeConfig(_Config, _HasFeatureConfigs,
         'simplex' uses d+1 parameters and thus scales better. For details see
         `tfl.lattice_lib.evaluate_with_simplex_interpolation` and
         `tfl.lattice_lib.evaluate_with_hypercube_interpolation`.
+      parameterization: The parameterization of the lattice function class to
+        use. A lattice function is uniquely determined by specifying its value
+        on every lattice vertex. A parameterization scheme is a mapping from a
+        vector of parameters to a multidimensional array of lattice vertex
+        values. It can be one of:
+          - String `'all_vertices'`: This is the "traditional" parameterization
+            that keeps one scalar parameter per lattice vertex where the mapping
+            is essentially the identity map. With this scheme, the number of
+            parameters scales exponentially with the number of inputs to the
+            lattice. The underlying lattice used will be a `tfl.layers.Lattice`
+            layer.
+          - String `'kronecker_factored'`: With this parameterization, for each
+            lattice input i we keep a collection of `num_terms` vectors each
+            having `feature_configs[0].lattice_size` entries (note that all
+            features must have the same lattice size). To obtain the tensor of
+            lattice vertex values, for `t=1,2,...,num_terms` we compute the
+            outer product of the `t'th` vector in each collection, multiply by a
+            per-term scale, and sum the resulting tensors. Finally, we add a
+            single shared bias parameter to each entry in the sum. With this
+            scheme, the number of parameters grows linearly with
+            `len(feature_configs)` (assuming lattice sizes and `num_terms` are
+            held constant). Currently, only monotonicity shape constraint and
+            bound constraint are supported for this scheme. Regularization is
+            not currently supported. The underlying lattice used will be a
+            `tfl.layers.KroneckerFactoredLattice` layer.
+      num_terms: The number of terms in a lattice using `'kronecker_factored'`
+        parameterization. Ignored if parameterization is set to
+        `'all_vertices'`.
       regularizer_configs: A list of `tfl.configs.RegularizerConfig` instances
         that apply global regularization.
       output_min: Lower bound constraint on the output of the model.
@@ -410,6 +478,14 @@ class CalibratedLatticeConfig(_Config, _HasFeatureConfigs,
           - String `'uniform'`: Output is initliazed uniformly in label range.
           - A list of numbers: To be used for initialization of the output
             lattice or output calibrator.
+      output_calibration_input_keypoints_type: One of "fixed" or
+        "learned_interior". If "learned_interior", keypoints are initialized to
+        the values in `pwl_calibration_input_keypoints` but then allowed to vary
+        during training, with the exception of the first and last keypoint
+        location which are fixed.
+      random_seed: Random seed to use for initialization of a lattice with
+        `'kronecker_factored'` parameterization. Ignored if parameterization is
+        set to `'all_vertices'`.
     """
     super(CalibratedLatticeConfig, self).__init__(locals())
 
@@ -454,7 +530,8 @@ class CalibratedLinearConfig(_Config, _HasFeatureConfigs,
                output_max=None,
                output_calibration=False,
                output_calibration_num_keypoints=10,
-               output_initialization='quantiles'):
+               output_initialization='quantiles',
+               output_calibration_input_keypoints_type='fixed'):
     """Initializes a `CalibratedLinearConfig` instance.
 
     Args:
@@ -480,6 +557,11 @@ class CalibratedLinearConfig(_Config, _HasFeatureConfigs,
           - String `'uniform'`: Output is initliazed uniformly in label range.
           - A list of numbers: To be used for initialization of the output
             lattice or output calibrator.
+      output_calibration_input_keypoints_type: One of "fixed" or
+        "learned_interior". If "learned_interior", keypoints are initialized to
+        the values in `pwl_calibration_input_keypoints` but then allowed to vary
+        during training, with the exception of the first and last keypoint
+        location which are fixed.
     """
     super(CalibratedLinearConfig, self).__init__(locals())
 
@@ -520,6 +602,7 @@ class AggregateFunctionConfig(_Config, _HasFeatureConfigs,
                middle_lattice_size=2,
                middle_calibration=False,
                middle_calibration_num_keypoints=10,
+               middle_calibration_input_keypoints_type='fixed',
                middle_monotonicity=None,
                middle_lattice_interpolation='hypercube',
                aggregation_lattice_interpolation='hypercube',
@@ -527,7 +610,8 @@ class AggregateFunctionConfig(_Config, _HasFeatureConfigs,
                output_max=None,
                output_calibration=False,
                output_calibration_num_keypoints=10,
-               output_initialization='uniform'):
+               output_initialization='uniform',
+               output_calibration_input_keypoints_type='fixed'):
     """Initializes an `AggregateFunctionConfig` instance.
 
     Args:
@@ -545,6 +629,11 @@ class AggregateFunctionConfig(_Config, _HasFeatureConfigs,
         the inputs to the middle lattice.
       middle_calibration_num_keypoints: Number of keypoints to use for the
         middle piecewise-linear calibration.
+      middle_calibration_input_keypoints_type: One of "fixed" or
+        "learned_interior". If "learned_interior", keypoints are initialized to
+        the values in `pwl_calibration_input_keypoints` but then allowed to vary
+        during training, with the exception of the first and last keypoint
+        location which are fixed.
       middle_monotonicity: Specifies if the middle calibrators should be
         monotonic, using 'increasing' or 1 to indicate increasing monotonicity,
         'decreasing' or -1 to indicate decreasing monotonicity, and 'none' or 0
@@ -573,6 +662,11 @@ class AggregateFunctionConfig(_Config, _HasFeatureConfigs,
           - String `'uniform'`: Output is initliazed uniformly in label range.
           - A list of numbers: To be used for initialization of the output
             lattice or output calibrator.
+      output_calibration_input_keypoints_type: One of "fixed" or
+        "learned_interior". If "learned_interior", keypoints are initialized to
+        the values in `pwl_calibration_input_keypoints` but then allowed to vary
+        during training, with the exception of the first and last keypoint
+        location which are fixed.
     """
     super(AggregateFunctionConfig, self).__init__(locals())
 
@@ -652,6 +746,7 @@ class FeatureConfig(_Config, _HasRegularizerConfigs):
                pwl_calibration_convexity=0,
                pwl_calibration_num_keypoints=10,
                pwl_calibration_input_keypoints='quantiles',
+               pwl_calibration_input_keypoints_type='fixed',
                pwl_calibration_clip_min=None,
                pwl_calibration_clip_max=None,
                pwl_calibration_clamp_min=False,
@@ -707,6 +802,11 @@ class FeatureConfig(_Config, _HasRegularizerConfigs):
           - String `'uniform'`: Input keypoints are uniformly spaced in feature
             range.
           - A list of numbers: Explicitly specifies the keypoints.
+      pwl_calibration_input_keypoints_type: One of "fixed" or
+        "learned_interior". If "learned_interior", keypoints are initialized to
+        the values in `pwl_calibration_input_keypoints` but then allowed to vary
+        during training, with the exception of the first and last keypoint
+        location which are fixed. Convexity can only be imposed with "fixed".
       pwl_calibration_clip_min: Input values are lower clipped by this value.
       pwl_calibration_clip_max: Input values are upper clipped by this value.
       pwl_calibration_clamp_min: for monotonic calibrators ensures that the
