@@ -73,6 +73,7 @@ import time
 
 from . import categorical_calibration_layer
 from . import configs
+from . import kronecker_factored_lattice_layer as kfll
 from . import lattice_layer
 from . import linear_layer
 from . import model_info
@@ -223,8 +224,13 @@ def _materialize_locally(tensors, max_elements=1e6):
       pass
     concatenated_tensors = {}
     for k in tensors:
-      concatenated_tensors[k] = np.concatenate(
-          [split[k] for split in splits if split[k].size > 0])
+      k_tensors = [split[k] for split in splits if split[k].size > 0]
+      # If k_tensors is empty, the np.concat call below would raise
+      # an exception below. So we test this condition here to provide
+      # a better error message.
+      if not k_tensors:
+        raise ValueError("Did not find any values for key: {}".format(k))
+      concatenated_tensors[k] = np.concatenate(k_tensors)
     return concatenated_tensors
 
 
@@ -252,7 +258,10 @@ def _finalize_keypoints(model_config, config, feature_columns,
   if ((config is None or config.is_chief) and
       not tf.io.gfile.exists(keypoints_filename)):
     with tf.Graph().as_default():
-      features, label = feature_analysis_input_fn()
+      ds = feature_analysis_input_fn()
+      if isinstance(ds, tf.data.Dataset):
+        ds = tf.compat.v1.data.make_one_shot_iterator(ds).get_next()
+      features, label = ds
       features = transform_features(features, feature_columns)
       features[_LABEL_FEATURE_NAME] = label
       features = _materialize_locally(features)
@@ -954,11 +963,16 @@ class CannedClassifier(estimator_lib.EstimatorV2):
         classification. Must be > 1.
       weight_column: A string or a `_NumericColumn` created by
         `tf.feature_column.numeric_column` defining feature column representing
-        weights. It is used to down weight or boost examples during training. It
-        will be multiplied by the loss of the example. If it is a string, it is
-        used as a key to fetch weight tensor from the `features`. If it is a
-        `_NumericColumn`, raw tensor is fetched by key `weight_column.key`, then
-        weight_column.normalizer_fn is applied on it to get weight tensor.
+        weights. It is only used by the estimator head to down weight or boost
+        examples during training. It will be multiplied by the loss of the
+        example. If it is a string, it is used as a key to fetch the weight
+        tensor from the `features` dictionary output of the input function.
+        If it is a `_NumericColumn`, a raw tensor is fetched by key
+        `weight_column.key`, then weight_column.normalizer_fn is applied on it
+        to get the weight tensor. Note that in both cases 'weight_column'
+        should *not* be a member of the 'feature_columns'  parameter
+        to the constructor since these will be used for both serving and
+        training.
       label_vocabulary: A list of strings represents possible label values. If
         given, labels must be string type and have any value in
         `label_vocabulary`. If it is not given, that means labels are already
@@ -1110,11 +1124,16 @@ class CannedRegressor(estimator_lib.EstimatorV2):
         (typically, these have shape `[batch_size, label_dimension]`).
       weight_column: A string or a `_NumericColumn` created by
         `tf.feature_column.numeric_column` defining feature column representing
-        weights. It is used to down weight or boost examples during training. It
-        will be multiplied by the loss of the example. If it is a string, it is
-        used as a key to fetch weight tensor from the `features`. If it is a
-        `_NumericColumn`, raw tensor is fetched by key `weight_column.key`, then
-        weight_column.normalizer_fn is applied on it to get weight tensor.
+        weights. It is only used by the estimator head to down weight or boost
+        examples during training. It will be multiplied by the loss of the
+        example. If it is a string, it is used as a key to fetch the weight
+        tensor from the `features` dictionary output of the input function.
+        If it is a `_NumericColumn`, a raw tensor is fetched by key
+        `weight_column.key`, then weight_column.normalizer_fn is applied on it
+        to get the weight tensor. Note that in both cases 'weight_column'
+        should *not* be a member of the 'feature_columns'  parameter
+        to the constructor since these will be used for both serving and
+        training.
       optimizer: An instance of `tf.Optimizer` used to train the model. Can also
         be a string (one of 'Adagrad', 'Adam', 'Ftrl', 'RMSProp', 'SGD'), or
         callable. Defaults to Adagrad optimizer.
@@ -1202,10 +1221,22 @@ def _create_feature_nodes(sess, ops, graph):
   # Extract list of features from the graph.
   # {FEATURES_SCOPE}/{feature_name}
   feature_nodes = {}
-  feature_op_re = '{}/(.*)'.format(FEATURES_SCOPE)
+  feature_op_re = r'^{}/(.*)'.format(re.escape(FEATURES_SCOPE))
   for (_, feature_name) in _match_op(ops, feature_op_re):
-    category_table_op = 'transform/{}_lookup/Const'.format(feature_name)
-    if category_table_op in ops:
+    is_categorical = False
+    vocabulary_list = None
+
+    # Check to see if there is a categorical mapping defined for this feature
+    # (i.e. lookup table for categorical feature column). Note that there can be
+    # at most one such mapping either under transform or transform_1 namespace.
+    # transform(_\d)?/{feature_name}_lookup/Const
+    category_table_re = r'transform(_\d)?/{}_lookup/Const'.format(
+        re.escape(feature_name))
+    for (category_table_op, _) in _match_op(ops, category_table_re):
+      if is_categorical:
+        raise ValueError(
+            'Model graph has multiple category mappings for feature {}'
+            .format(feature_name))
       is_categorical = True
       vocabulary_list = sess.run(
           graph.get_operation_by_name(category_table_op).outputs[0])
@@ -1214,9 +1245,6 @@ def _create_feature_nodes(sess, ops, graph):
           str(x.decode()) if isinstance(x, bytes) else str(x)
           for x in vocabulary_list
       ]
-    else:
-      is_categorical = False
-      vocabulary_list = None
 
     feature_node = model_info.InputFeatureNode(
         name=feature_name,
@@ -1301,9 +1329,23 @@ def _create_pwl_calibration_nodes(sess, ops, graph, feature_nodes):
          graph.get_operation_by_name(kernel_op).outputs[0]))
     output_keypoints = np.cumsum(kernel, axis=0)
 
-    # Add the last keypoint to the keypoint list.
+    # For calibrators with fixed input keypoints, the shape of 'keypoints'
+    # and 'lengths' is (num_keypoints - 1,). For calibrators with keypoint_type
+    # set to learned_interior the shape is (num_units, num_keypoints - 1).
+    # Change the shape to (num_units, num_keypoints - 1).
+    # If needed, repeat the keypoints for each output unit.
+    if keypoints.ndim == 1:
+      keypoints = np.expand_dims(keypoints, axis=0)
+      lengths = np.expand_dims(lengths, axis=0)
+    if keypoints.shape[0] == 1:
+      keypoints = np.tile(keypoints, [output_keypoints.shape[1], 1])
+      lengths = np.tile(lengths, [output_keypoints.shape[1], 1])
+
+    # Add the last keypoint to the keypoint list. The resulting shape is
+    # (num_units, num_keypoints).
     # TODO: handle cyclic PWL layers.
-    input_keypoints = np.append(keypoints, keypoints[-1] + lengths[-1])
+    input_keypoints = np.concatenate(
+        [keypoints, keypoints[:, -1:] + lengths[:, -1:]], axis=-1)
 
     # Get missing/default input value if present:
     # {CALIB_LAYER_NAME}_{feature_name}/{MISSING_INPUT_VALUE_NAME}
@@ -1335,7 +1377,7 @@ def _create_pwl_calibration_nodes(sess, ops, graph, feature_nodes):
     for calibration_output_idx in range(output_keypoints.shape[1]):
       pwl_calibration_node = model_info.PWLCalibrationNode(
           input_node=feature_nodes[feature_name],
-          input_keypoints=input_keypoints,
+          input_keypoints=input_keypoints[calibration_output_idx, :],
           output_keypoints=output_keypoints[:, calibration_output_idx],
           default_input=default_input,
           default_output=(None if default_output is None else
@@ -1433,8 +1475,251 @@ def _create_lattice_nodes(sess, ops, graph, submodel_input_nodes):
   return lattice_nodes
 
 
-def _create_rtl_lattice_nodes(sess, ops, graph, calibration_nodes_map):
-  """Returns a map from lattice_submodel_index to LatticeNode."""
+def _create_kronecker_factored_lattice_nodes(sess, ops, graph,
+                                             submodel_input_nodes):
+  """Returns a map from submodel_idx to KroneckerFactoredLatticeNode."""
+  kfl_nodes = {}
+  # KroneckerFactoredLattice kernel weights.
+  # {KFL_LAYER_NAME}_{submodel_idx}/{KFL_KERNEL_NAME}
+  kfl_kernel_op_re = '^{}_(.*)/{}/Read/ReadVariableOp$'.format(
+      premade_lib.KFL_LAYER_NAME,
+      kfll.KFL_KERNEL_NAME,
+  )
+  for kfl_kernel_op, submodel_idx in _match_op(ops, kfl_kernel_op_re):
+    kfl_kernel = sess.run(
+        graph.get_operation_by_name(kfl_kernel_op).outputs[0]).flatten()
+
+    # KroneckerFactoredLattice scale.
+    # {KFL_LAYER_NAME}_{submodel_idx}/{KFL_SCALE_NAME}
+    kfl_scale_op_name = '{}_{}/{}/Read/ReadVariableOp'.format(
+        premade_lib.KFL_LAYER_NAME, submodel_idx, kfll.KFL_SCALE_NAME)
+    kfl_scale = sess.run(
+        graph.get_operation_by_name(kfl_scale_op_name).outputs[0]).flatten()
+
+    # KroneckerFactoredLattice bias.
+    # {KFL_LAYER_NAME}_{submodel_idx}/{KFL_BIAS_NAME}
+    kfl_bias_op_name = '{}_{}/{}/Read/ReadVariableOp'.format(
+        premade_lib.KFL_LAYER_NAME, submodel_idx, kfll.KFL_BIAS_NAME)
+    kfl_bias = sess.run(
+        graph.get_operation_by_name(kfl_bias_op_name).outputs[0]).flatten()
+
+    # Lattice sizes.
+    # {KFL_LAYER_NAME}_{submodel_idx}/{LATTICE_SIZES_NAME}
+    lattice_sizes_op_name = '{}_{}/{}'.format(premade_lib.KFL_LAYER_NAME,
+                                              submodel_idx,
+                                              kfll.LATTICE_SIZES_NAME)
+    lattice_sizes = sess.run(
+        graph.get_operation_by_name(
+            lattice_sizes_op_name).outputs[0])
+
+    # Units.
+    # {KFL_LAYER_NAME}_{submodel_idx}/{UNITS_NAME}
+    units_op_name = '{}_{}/{}'.format(premade_lib.KFL_LAYER_NAME,
+                                      submodel_idx, kfll.UNITS_NAME)
+    units = sess.run(
+        graph.get_operation_by_name(units_op_name).outputs[0])
+
+    # Dims.
+    # {KFL_LAYER_NAME}_{submodel_idx}/{DIMS_NAME}
+    dims_op_name = '{}_{}/{}'.format(premade_lib.KFL_LAYER_NAME,
+                                     submodel_idx, kfll.DIMS_NAME)
+    dims = sess.run(
+        graph.get_operation_by_name(dims_op_name).outputs[0])
+
+    # Num terms.
+    # {KFL_LAYER_NAME}_{submodel_idx}/{NUM_TERMS_NAME}
+    num_terms_op_name = '{}_{}/{}'.format(premade_lib.KFL_LAYER_NAME,
+                                          submodel_idx, kfll.NUM_TERMS_NAME)
+    num_terms = sess.run(
+        graph.get_operation_by_name(num_terms_op_name).outputs[0])
+
+    # Shape the flat weights, scale, and bias parameters based on the calculated
+    # lattice_sizes, units, dims, and num_terms.
+    weights = np.reshape(kfl_kernel,
+                         (1, lattice_sizes, units * dims, num_terms))
+    scale = np.reshape(kfl_scale, (units, num_terms))
+    bias = np.reshape(kfl_bias, (units))
+
+    # Sort input nodes by input index.
+    input_nodes = [
+        node for _, node in sorted(submodel_input_nodes[submodel_idx])
+    ]
+
+    kfl_node = model_info.KroneckerFactoredLatticeNode(
+        input_nodes=input_nodes, weights=weights, scale=scale, bias=bias)
+    kfl_nodes[submodel_idx] = kfl_node
+  return kfl_nodes
+
+
+def _create_rtl_submodel_kronecker_factored_lattice_nodes(
+    sess, ops, graph, flattened_calibration_nodes, submodel_idx, submodel_key):
+  """Returns next key and map from key+unit to KroneckerFactoredLatticeNode."""
+  submodel_kfl_nodes = {}
+  # KroneckerFactoredLattice kernel weights
+  # {RTL_LAYER_NAME}_{submodel_idx}/
+  # {RTL_KFL_NAME}_{monotonicities}/{KFL_KERNEL_NAME}
+  kfl_kernel_op_re = '^{}_{}/{}_(.*)/{}/Read/ReadVariableOp$'.format(
+      premade_lib.RTL_LAYER_NAME,
+      submodel_idx,
+      rtl_layer.RTL_KFL_NAME,
+      kfll.KFL_KERNEL_NAME,
+  )
+  for kfl_kernel_op, monotonicities in _match_op(ops, kfl_kernel_op_re):
+    kfl_kernel = sess.run(
+        graph.get_operation_by_name(kfl_kernel_op).outputs[0]).flatten()
+
+    # KroneckerFactoredLattice scale.
+    # {RTL_LAYER_NAME}_{submodel_idx}/
+    # {RTL_KFL_NAME}_{monotonicities}/{KFL_SCALE_NAME}
+    kfl_scale_op_name = '{}_{}/{}_{}/{}/Read/ReadVariableOp'.format(
+        premade_lib.RTL_LAYER_NAME,
+        submodel_idx,
+        rtl_layer.RTL_KFL_NAME,
+        monotonicities,
+        kfll.KFL_SCALE_NAME,
+    )
+    kfl_scale = sess.run(
+        graph.get_operation_by_name(kfl_scale_op_name).outputs[0]).flatten()
+
+    # KroneckerFactoredLattice bias.
+    # {RTL_LAYER_NAME}_{submodel_idx}/
+    # {RTL_KFL_NAME}_{monotonicities}/{KFL_BIAS_NAME}
+    kfl_bias_op_name = '{}_{}/{}_{}/{}/Read/ReadVariableOp'.format(
+        premade_lib.RTL_LAYER_NAME,
+        submodel_idx,
+        rtl_layer.RTL_KFL_NAME,
+        monotonicities,
+        kfll.KFL_BIAS_NAME,
+    )
+    kfl_bias = sess.run(
+        graph.get_operation_by_name(kfl_bias_op_name).outputs[0]).flatten()
+
+    # Lattice sizes.
+    # {RTL_LAYER_NAME}_{submodel_idx}/
+    # {RTL_KFL_NAME}_{monotonicities}/{LATTICE_SIZES_NAME}
+    lattice_sizes_op_name = '{}_{}/{}_{}/{}'.format(
+        premade_lib.RTL_LAYER_NAME,
+        submodel_idx,
+        rtl_layer.RTL_KFL_NAME,
+        monotonicities,
+        kfll.LATTICE_SIZES_NAME,
+    )
+    lattice_sizes = sess.run(
+        graph.get_operation_by_name(lattice_sizes_op_name).outputs[0])
+
+    # Dims.
+    # {RTL_LAYER_NAME}_{submodel_idx}/
+    # {RTL_KFL_NAME}_{monotonicities}/{DIMS_NAME}
+    dims_op_name = '{}_{}/{}_{}/{}'.format(
+        premade_lib.RTL_LAYER_NAME,
+        submodel_idx,
+        rtl_layer.RTL_KFL_NAME,
+        monotonicities,
+        kfll.DIMS_NAME,
+    )
+    dims = sess.run(graph.get_operation_by_name(dims_op_name).outputs[0])
+
+    # Num terms.
+    # {RTL_LAYER_NAME}_{submodel_idx}/
+    # {RTL_KFL_NAME}_{monotonicities}/{NUM_TERMS_NAME}
+    num_terms_op_name = '{}_{}/{}_{}/{}'.format(
+        premade_lib.RTL_LAYER_NAME,
+        submodel_idx,
+        rtl_layer.RTL_KFL_NAME,
+        monotonicities,
+        kfll.NUM_TERMS_NAME,
+    )
+    num_terms = sess.run(
+        graph.get_operation_by_name(num_terms_op_name).outputs[0])
+
+    # inputs_for_units
+    # {RTL_LAYER_NAME}_{submodel_index}/
+    # {INPUTS_FOR_UNITS_PREFIX}_{monotonicities}
+    inputs_for_units_op_name = '{}_{}/{}_{}'.format(
+        premade_lib.RTL_LAYER_NAME, submodel_idx,
+        rtl_layer.INPUTS_FOR_UNITS_PREFIX, monotonicities)
+    inputs_for_units = sess.run(
+        graph.get_operation_by_name(inputs_for_units_op_name).outputs[0])
+
+    # Make a unique kfl for each unit.
+    units = inputs_for_units.shape[0]
+    for i in range(units):
+      # Shape the flat weights, scale, and bias parameters based on the
+      # calculated lattice_sizes, units, dims, and num_terms.
+      weights = np.reshape(kfl_kernel,
+                           (1, lattice_sizes, units * dims, num_terms))
+      scale = np.reshape(kfl_scale, (units, num_terms))
+      bias = np.reshape(kfl_bias, (units))
+
+      # Gather input nodes for lattice node.
+      indices = inputs_for_units[i]
+      input_nodes = [flattened_calibration_nodes[index] for index in indices]
+
+      kfl_node = model_info.KroneckerFactoredLatticeNode(
+          input_nodes=input_nodes, weights=weights, scale=scale, bias=bias)
+      submodel_kfl_nodes[submodel_key] = kfl_node
+      submodel_key += 1
+  return submodel_key, submodel_kfl_nodes
+
+
+def _create_rtl_submodel_lattice_nodes(sess, ops, graph,
+                                       flattened_calibration_nodes,
+                                       submodel_idx, submodel_key):
+  """Returns next key and map from key+unit to LatticeNode."""
+  submodel_lattice_nodes = {}
+  # Lattice kernel weights.
+  # {RTL_LAYER_NAME}_{submodel_idx}/
+  # {RTL_LATTICE_NAME}_{monotonicities}/{LATTICE_KERNEL_NAME}
+  lattice_kernel_op_re = '^{}_{}/{}_(.*)/{}/Read/ReadVariableOp$'.format(
+      premade_lib.RTL_LAYER_NAME,
+      submodel_idx,
+      rtl_layer.RTL_LATTICE_NAME,
+      lattice_layer.LATTICE_KERNEL_NAME,
+  )
+  for lattice_kernel_op, monotonicities in _match_op(ops, lattice_kernel_op_re):
+    lattice_kernel = sess.run(
+        graph.get_operation_by_name(lattice_kernel_op).outputs[0])
+
+    # Lattice sizes.
+    # {RTL_LAYER_NAME}_{submodel_idx}/
+    # {RTL_LATTICE_NAME}_{monotonicities}/{LATTICE_SIZES_NAME}
+    lattice_sizes_op_name = '{}_{}/{}_{}/{}'.format(
+        premade_lib.RTL_LAYER_NAME, submodel_idx, rtl_layer.RTL_LATTICE_NAME,
+        monotonicities, lattice_layer.LATTICE_SIZES_NAME)
+    lattice_sizes = sess.run(
+        graph.get_operation_by_name(
+            lattice_sizes_op_name).outputs[0]).flatten()
+
+    # inputs_for_units
+    # {RTL_LAYER_NAME}_{submodel_index}/
+    # {INPUTS_FOR_UNITS_PREFIX}_{monotonicities}
+    inputs_for_units_op_name = '{}_{}/{}_{}'.format(
+        premade_lib.RTL_LAYER_NAME, submodel_idx,
+        rtl_layer.INPUTS_FOR_UNITS_PREFIX, monotonicities)
+    inputs_for_units = sess.run(
+        graph.get_operation_by_name(inputs_for_units_op_name).outputs[0])
+
+    # Make a unique lattice for each unit.
+    units = inputs_for_units.shape[0]
+    for i in range(units):
+      # Shape the flat lattice parameters based on the calculated lattice
+      # sizes.
+      weights = np.reshape(lattice_kernel[:, i], lattice_sizes)
+
+      # Gather input nodes for lattice node.
+      indices = inputs_for_units[i]
+      input_nodes = [flattened_calibration_nodes[index] for index in indices]
+
+      lattice_node = model_info.LatticeNode(
+          input_nodes=input_nodes, weights=weights)
+      submodel_lattice_nodes[submodel_key] = lattice_node
+      submodel_key += 1
+  return submodel_key, submodel_lattice_nodes
+
+
+def _create_rtl_lattice_nodes(sess, ops, graph, calibration_nodes_map,
+                              kronecker_factored):
+  """Returns a map from lattice_submodel_index to lattice type Node."""
   lattice_nodes = {}
   lattice_submodel_index = 0
   # Feature name in concat op.
@@ -1458,57 +1743,14 @@ def _create_rtl_lattice_nodes(sess, ops, graph, calibration_nodes_map):
     for feature_name in names_in_flattened_order:
       flattened_calibration_nodes.extend(calibration_nodes_map[feature_name])
 
-    # Lattice kernel weights.
-    # {RTL_LAYER_NAME}_{submodel_idx}/
-    # {RTL_LATTICE_NAME}_{monotonicities}/{LATTICE_KERNEL_NAME}
-    lattice_kernel_op_re = '^{}_{}/{}_(.*)/{}/Read/ReadVariableOp$'.format(
-        premade_lib.RTL_LAYER_NAME,
-        submodel_idx,
-        rtl_layer.RTL_LATTICE_NAME,
-        lattice_layer.LATTICE_KERNEL_NAME,
-    )
-    for lattice_kernel_op, monotonicities in _match_op(ops,
-                                                       lattice_kernel_op_re):
-      # Lattice kernel weights.
-      lattice_kernel = sess.run(
-          graph.get_operation_by_name(lattice_kernel_op).outputs[0])
-
-      # Lattice sizes.
-      # {RTL_LAYER_NAME}_{submodel_idx}/
-      # {RTL_LATTICE_NAME}_{monotonicities}/{LATTICE_SIZES_NAME}
-      lattice_sizes_op_name = '{}_{}/{}_{}/{}'.format(
-          premade_lib.RTL_LAYER_NAME, submodel_idx, rtl_layer.RTL_LATTICE_NAME,
-          monotonicities, lattice_layer.LATTICE_SIZES_NAME)
-
-      lattice_sizes = sess.run(
-          graph.get_operation_by_name(
-              lattice_sizes_op_name).outputs[0]).flatten()
-
-      # inputs_for_units
-      # {RTL_LAYER_NAME}_{submodel_index}/
-      # {INPUTS_FOR_UNITS_PREFIX}_{monotonicities}
-      inputs_for_units_op_name = '{}_{}/{}_{}'.format(
-          premade_lib.RTL_LAYER_NAME, submodel_idx,
-          rtl_layer.INPUTS_FOR_UNITS_PREFIX, monotonicities)
-
-      inputs_for_units = sess.run(
-          graph.get_operation_by_name(inputs_for_units_op_name).outputs[0])
-
-      # Make a unique lattice for each unit.
-      units = inputs_for_units.shape[0]
-      for i in range(units):
-        # Shape the flat lattice parameters based on the calculated lattice
-        # sizes.
-        weights = np.reshape(lattice_kernel[:, i], lattice_sizes)
-
-        # Gather input nodes for lattice node.
-        indices = inputs_for_units[i]
-        input_nodes = [flattened_calibration_nodes[index] for index in indices]
-
-        lattice_node = model_info.LatticeNode(
-            input_nodes=input_nodes, weights=weights)
-        lattice_nodes[lattice_submodel_index] = lattice_node
-        lattice_submodel_index += 1
+    if kronecker_factored:
+      node_fn = _create_rtl_submodel_kronecker_factored_lattice_nodes
+    else:
+      node_fn = _create_rtl_submodel_lattice_nodes
+    lattice_submodel_index, submodel_lattice_nodes = node_fn(
+        sess, ops, graph, flattened_calibration_nodes, submodel_idx,
+        lattice_submodel_index)
+    lattice_nodes.update(submodel_lattice_nodes)
   return lattice_nodes
 
 
@@ -1595,6 +1837,7 @@ def _create_output_calibration_node(sess, ops, graph, input_node):
   return output_calibration_node
 
 
+# TODO: add support for KFL in RTL Layer
 def get_model_graph(saved_model_path, tag='serve'):
   """Returns all layers and parameters used in a saved model as a graph.
 
@@ -1666,12 +1909,24 @@ def get_model_graph(saved_model_path, tag='serve'):
     submodel_output_nodes.update(lattice_nodes)
     nodes.extend(lattice_nodes.values())
 
+    # Ensemble Kronecker Factored Lattice nodes.
+    kfl_nodes = _create_kronecker_factored_lattice_nodes(
+        sess, ops, graph, submodel_input_nodes)
+    submodel_output_nodes.update(kfl_nodes)
+    nodes.extend(kfl_nodes.values())
+
     # RTL Lattice nodes.
-    rtl_lattice_nodes = _create_rtl_lattice_nodes(sess, ops, graph,
-                                                  calibration_nodes_map)
+    rtl_lattice_nodes = _create_rtl_lattice_nodes(
+        sess, ops, graph, calibration_nodes_map, kronecker_factored=False)
     submodel_output_nodes.update(rtl_lattice_nodes)
     nodes.extend(rtl_lattice_nodes.values())
 
+    # RTL Kronecker Factored Lattice nodes.
+    kfl_rtl_nodes = _create_rtl_lattice_nodes(
+        sess, ops, graph, calibration_nodes_map, kronecker_factored=True)
+    submodel_output_nodes.update(kfl_rtl_nodes)
+    nodes.extend(kfl_rtl_nodes.values())
+
     # Output combination node.
     model_output_node = _create_output_combination_node(sess, ops, graph,
                                                         submodel_output_nodes)
