@@ -40,7 +40,8 @@ class CannedEstimatorsTest(parameterized.TestCase, tf.test.TestCase):
 
     # UCI Statlog (Heart) dataset.
     heart_csv_file = tf.keras.utils.get_file(
-        'heart.csv', 'http://storage.googleapis.com/applied-dl/heart.csv')
+        'heart.csv',
+        'http://storage.googleapis.com/download.tensorflow.org/data/heart.csv')
     heart_df = pd.read_csv(heart_csv_file)
     heart_target = heart_df.pop('target')
     heart_train_size = int(len(heart_df) * 0.8)
@@ -258,14 +259,22 @@ class CannedEstimatorsTest(parameterized.TestCase, tf.test.TestCase):
     tf.keras.backend.clear_session()
     tf.compat.v1.reset_default_graph()
 
-  def _GetInputFn(self, x, y, num_epochs=1, batch_size=100):
-    return tf.compat.v1.estimator.inputs.pandas_input_fn(
-        x=x,
-        y=y,
-        batch_size=batch_size,
-        shuffle=False,
-        num_epochs=num_epochs,
-        num_threads=1)
+  def _GetInputFn(self, x, y, num_epochs=1, batch_size=100, tfds=False):
+    if tfds:
+
+      def _input_fn():
+        return tf.data.Dataset.from_tensor_slices(
+            (x.to_dict('list'), y.values)).batch(batch_size).repeat(num_epochs)
+
+      return _input_fn
+    else:
+      return tf.compat.v1.estimator.inputs.pandas_input_fn(
+          x=x,
+          y=y,
+          batch_size=batch_size,
+          shuffle=False,
+          num_epochs=num_epochs,
+          num_threads=1)
 
   def _GetHeartTrainInputFn(self, **kwargs):
     return self._GetInputFn(self.heart_train_x, self.heart_train_y, **kwargs)
@@ -594,17 +603,18 @@ class CannedEstimatorsTest(parameterized.TestCase, tf.test.TestCase):
     self.assertLess(results['average_loss'], average_loss)
 
   @parameterized.parameters(
-      ('random', 5, 6, False, True),
-      ('random', 4, 5, True, False),
-      ('rtl_layer', 5, 6, False, True),
-      ('rtl_layer', 4, 5, True, False),
+      ('random', 5, 6, 'all_vertices', False, True),
+      ('random', 4, 5, 'kronecker_factored', True, False),
+      ('rtl_layer', 5, 6, 'kronecker_factored', False, True),
+      ('rtl_layer', 4, 5, 'all_vertices', True, False),
   )
   def testCalibratedLatticeEnsembleModelInfo(self, lattices, num_lattices,
-                                             lattice_rank, separate_calibrators,
+                                             lattice_rank, parameterization,
+                                             separate_calibrators,
                                              output_calibration):
     self._ResetAllBackends()
     feature_configs = copy.deepcopy(self.heart_feature_configs)
-    if lattices == 'rtl_layer':
+    if lattices == 'rtl_layer' or parameterization == 'kronecker_factored':
       # RTL Layer only supports monotonicity and bound constraints.
       for feature_config in feature_configs:
         feature_config.lattice_size = 2
@@ -617,6 +627,7 @@ class CannedEstimatorsTest(parameterized.TestCase, tf.test.TestCase):
         lattices=lattices,
         num_lattices=num_lattices,
         lattice_rank=lattice_rank,
+        parameterization=parameterization,
         separate_calibrators=separate_calibrators,
         output_calibration=output_calibration,
     )
@@ -705,11 +716,33 @@ class CannedEstimatorsTest(parameterized.TestCase, tf.test.TestCase):
     for lattice, expected_lattice in zip(lattices, expected_lattices):
       self.assertCountEqual(lattice, expected_lattice)
 
+  @parameterized.parameters((True,), (False,))
+  def testDatasetAPI(self, tfds):
+    self._ResetAllBackends()
+    feature_columns = self.heart_feature_columns
+    feature_configs = self.heart_feature_configs
+    model_config = configs.CalibratedLinearConfig(
+        feature_configs=feature_configs)
+    estimator = estimators.CannedClassifier(
+        feature_columns=feature_columns,
+        model_config=model_config,
+        feature_analysis_input_fn=self._GetHeartTrainInputFn(
+            num_epochs=1, tfds=tfds),
+        optimizer=tf.keras.optimizers.Adam(0.01))
+    estimator.train(
+        input_fn=self._GetHeartTrainInputFn(num_epochs=200, tfds=tfds))
+    results = estimator.evaluate(input_fn=self._GetHeartTestInputFn(tfds=tfds))
+    logging.info('Calibrated linear classifier results:')
+    logging.info(results)
+    self.assertGreater(results['auc'], 0.7)
+
   @parameterized.parameters(
-      ('linear', True),
-      ('lattice', False),
+      ('linear', None, True),
+      ('lattice', 'all_vertices', False),
+      ('lattice', 'kronecker_factored', False),
   )
-  def testCalibratedModelInfo(self, model_type, output_calibration):
+  def testCalibratedModelInfo(self, model_type, parameterization,
+                              output_calibration):
     self._ResetAllBackends()
     if model_type == 'linear':
       model_config = configs.CalibratedLinearConfig(
@@ -717,8 +750,18 @@ class CannedEstimatorsTest(parameterized.TestCase, tf.test.TestCase):
           output_calibration=output_calibration,
       )
     else:
+      feature_configs = copy.deepcopy(self.heart_feature_configs)
+      if parameterization == 'kronecker_factored':
+        # RTL Layer only supports monotonicity and bound constraints.
+        for feature_config in feature_configs:
+          feature_config.lattice_size = 2
+          feature_config.unimodality = 'none'
+          feature_config.reflects_trust_in = None
+          feature_config.dominates = None
+          feature_config.regularizer_configs = None
       model_config = configs.CalibratedLatticeConfig(
-          feature_configs=self.heart_feature_configs,
+          feature_configs=feature_configs,
+          parameterization=parameterization,
           output_calibration=output_calibration,
       )
     estimator = estimators.CannedClassifier(
