@@ -149,12 +149,26 @@ def evaluate_with_hypercube_interpolation(inputs, scale, bias, kernel, units,
   return results
 
 
-def random_monotonic_initializer(shape,
-                                 monotonicities,
-                                 init_min=0.5,
-                                 init_max=1.5,
-                                 dtype=tf.float32,
-                                 seed=None):
+def default_init_params(output_min, output_max):
+  """Returns default initialization bounds depending on layer output bounds.
+
+  Args:
+    output_min: None or minimum layer output.
+    output_max: None or maximum layer output.
+  """
+  if output_min is None and output_max is None:
+    return 0.5, 1.5
+  else:
+    return 0.0, 1.0
+
+
+def kfl_random_monotonic_initializer(shape,
+                                     scale,
+                                     monotonicities,
+                                     init_min=0.5,
+                                     init_max=1.5,
+                                     dtype=tf.float32,
+                                     seed=None):
   """Returns a uniformly random sampled monotonic weight tensor.
 
   - The uniform random monotonic function will initilaize the lattice parameters
@@ -165,6 +179,7 @@ def random_monotonic_initializer(shape,
   Args:
     shape: Shape of weights to initialize. Must be: `(1, lattice_sizes, units *
       dims, num_terms)`.
+    scale: Scale variable of shape: `(units, num_terms)`.
     monotonicities: None or list or tuple of length dims of elements of {0,1}
       which represents monotonicity constraints per dimension. 1 stands for
       increasing (non-decreasing in fact), 0 for no monotonicity constraints.
@@ -175,7 +190,7 @@ def random_monotonic_initializer(shape,
 
   Returns:
     Kronecker-Factored Lattice weights tensor of shape:
-      `(1, lattice_sizes, units * dims, num_terms)`.
+    `(1, lattice_sizes, units * dims, num_terms)`.
   """
   # Sample from the uniform distribution.
   weights = tf.random.uniform(
@@ -186,37 +201,87 @@ def random_monotonic_initializer(shape,
     _, lattice_sizes, units_times_dims, num_terms = shape
     if units_times_dims % dims != 0:
       raise ValueError(
-          "len(monotonicities) is {}, which does not evenly divide shape[2]"
+          "len(monotonicities) is {}, which does not evenly divide shape[2]."
           "len(monotonicities) should be equal to `dims`, and shape[2] "
           "should be equal to units * dims.".format(dims))
     units = units_times_dims // dims
     weights = tf.reshape(weights, [-1, lattice_sizes, units, dims, num_terms])
+    # Make all dimensions monotonically increasing with respect to the sign of
+    # scale.
+    direction = tf.expand_dims(tf.sign(scale), axis=1)
     # Now we can unstack each dimension.
-    weights = tf.unstack(weights, axis=3)
+    weights = tf.unstack(direction * weights, axis=3)
     monotonic_weights = [
         tf.sort(weight, axis=1) if monotonicity else weight
         for monotonicity, weight in zip(monotonicities, weights)
     ]
     # Restack, reshape, and return weights
     weights = tf.stack(monotonic_weights, axis=3)
-    weights = tf.reshape(weights, shape)
+    weights = tf.reshape(direction * weights, shape)
   return weights
 
 
-def scale_initializer(units, num_terms):
-  """Initializes scale to alternate between 1 and -1 for each term.
+def scale_initializer(units, num_terms, output_min, output_max):
+  """Initializes scale depending on output_min and output_max.
+
+  If both output_min and output_max are set, scale is initialized to half their
+  difference, alternating signs for each term. If only output_min is set, scale
+  is initialized to 1 for each term. If only output_max is set, scale is
+  initialized to -1 for each term. Otherwise scale is initialized to alternate
+  between 1 and -1 for each term.
 
   Args:
-    units: Output dimension of the layer. Each of units scale will be
-      initialized identically.
+    units: Output dimension of the layer. Each unit's scale will be initialized
+      identically.
     num_terms: Number of independently trained submodels per unit, the outputs
       of which are averaged to get the final output.
+    output_min: None or minimum layer output.
+    output_max: None or maximum layer output.
 
   Returns:
     Kronecker-Factored Lattice scale of shape: `(units, num_terms)`.
   """
+  if output_min is not None and output_max is None:
+    return np.ones([units, num_terms])
+  if output_min is None and output_max is not None:
+    return -np.ones([units, num_terms])
+  # Both or neither bounds are set, so we alternate sign.
   signs = (np.arange(num_terms) % -2) * 2 + 1
-  return np.tile(signs, [units, 1])
+  scale = np.tile(signs, [units, 1])
+  if output_min is not None and output_max is not None:
+    scale = scale * ((output_max - output_min) / 2.0)
+  return scale
+
+
+def bias_initializer(units, output_min, output_max, dtype=tf.float32):
+  """Initializes bias depending on output_min and output_max.
+
+  If both output_min and output_max are set, bias is initialized to their
+  average. If only output_min is set, bias is initialized to output_min. If only
+  output_max is set, bias is initialized to output_max. Otherwise bias is
+  initialized to zeros.
+
+  Args:
+    units: Output dimension of the layer. Each of units bias will be initialized
+      identically.
+    output_min: None or minimum layer output.
+    output_max: None or maximum layer output.
+    dtype: dtype
+
+  Returns:
+    Kronecker-Factored Lattice bias of shape: `(units)`.
+  """
+  if output_min is not None and output_max is not None:
+    return tf.constant(
+        (output_min + output_max) / 2.0, shape=[units], dtype=dtype)
+  elif output_min is not None:
+    return tf.constant(output_min, shape=[units], dtype=dtype)
+  elif output_max is not None:
+    # In this case, weights will be nonnegative and scale will be nonpositive so
+    # we add output_max to interpolation output to achieve proper bound.
+    return tf.constant(output_max, shape=[units], dtype=dtype)
+  else:
+    return tf.zeros(shape=[units], dtype=dtype)
 
 
 def _approximately_project_monotonicity(weights, units, scale, monotonicities):
@@ -278,7 +343,56 @@ def _approximately_project_monotonicity(weights, units, scale, monotonicities):
   return weights
 
 
-def finalize_constraints(weights, units, scale, monotonicities):
+def _approximately_project_bounds(weights, units, output_min, output_max):
+  """Approximately projects to strictly meet bound constraints.
+
+  For more details, see _approximately_project_bounds in lattice_lib.py.
+
+  Args:
+    weights: Tensor with weights of shape `(1, lattice_sizes, units * dims,
+      num_terms)`.
+    units: Number of units per input dimension.
+    output_min: None or minimum layer output.
+    output_max: None or maximum layer output.
+
+  Returns:
+    Tensor with projected weights matching shape of input weights.
+  """
+  if output_min is None and output_max is None:
+    return weights
+
+  # We project by the dims'th root projection factor of the weights, ultimately
+  # projecting each term into the range [-1,1], but only if both output_min and
+  # output_max are specified. Otherwise, we restrict the weights to be
+  # nonnegative and the interpolation will do a final shift to respect the
+  # one-sided bound.
+  if output_min is not None and output_max is not None:
+    # Recall that w.shape is (1, lattice_sizes, units * dims, num_terms).
+    weights_shape = weights.get_shape().as_list()
+    _, lattice_sizes, units_times_dims, num_terms = weights_shape
+    assert units_times_dims % units == 0
+    dims = units_times_dims // units
+    weights = tf.reshape(weights, [-1, lattice_sizes, units, dims, num_terms])
+    max_keypoint_values = tf.reduce_max(tf.abs(weights), axis=1, keepdims=True)
+    max_output_value = tf.reduce_prod(
+        max_keypoint_values, axis=3, keepdims=True)
+    full_projection_factor = tf.maximum(max_output_value, 1.0)
+    individual_projection_factor = tf.pow(full_projection_factor, 1.0 / dims)
+    weights = weights / individual_projection_factor
+    # We must reshape to get our final projected weights.
+    weights = tf.reshape(weights, weights_shape)
+  else:
+    weights = tf.maximum(weights, 0)
+
+  return weights
+
+
+# Note: this function must not depend on the result of projecting scale.
+# Currently this function depends on the sign of scale, but the scale projection
+# will not flip the sign of scale (only make it 0 in the worse case), which will
+# not cause any issues.
+def finalize_weight_constraints(weights, units, scale, monotonicities,
+                                output_min, output_max):
   """Approximately projects weights to strictly satisfy all constraints.
 
   This projeciton guarantees that constraints are strictly met, but it is not
@@ -296,30 +410,72 @@ def finalize_constraints(weights, units, scale, monotonicities):
     monotonicities: List or tuple of length dims of elements of {0,1} which
       represents monotonicity constraints per dimension. 1 stands for increasing
       (non-decreasing in fact), 0 for no monotonicity constraints.
+    output_min: None or minimum layer output.
+    output_max: None or maximum layer output.
 
   Returns:
     Projected weights tensor of same shape as `weights`.
   """
-  if utils.count_non_zeros(monotonicities) == 0:
-    return weights
+  if utils.count_non_zeros(monotonicities) > 0:
+    # TODO: in the case of only one monotonic dimension, we only have to
+    # constrain the non-monotonic dimensions to be positive.
+    # There must be monotonicity constraints, so we need all nonnegative
+    # weights.
+    weights = tf.maximum(weights, 0)
+    weights = _approximately_project_monotonicity(
+        weights=weights,
+        units=units,
+        scale=scale,
+        monotonicities=monotonicities)
 
-  # TODO: in the case of only one monotonic dimension, we only have to
-  # constrain the non-monotonic dimensions to be positive.
-  # There must be monotonicity constraints, so we need all positive weights.
-  weights = tf.maximum(weights, 0)
-
-  # Project monotonicity constraints.
-  weights = _approximately_project_monotonicity(weights, units, scale,
-                                                monotonicities)
+  if output_min is not None or output_max is not None:
+    weights = _approximately_project_bounds(
+        weights=weights,
+        units=units,
+        output_min=output_min,
+        output_max=output_max)
 
   return weights
 
 
+# Note: we cannot rely on the weights projection occuring always before or
+# always after the scale projection, so this function must not result in a
+# projection that would ultimately change the results of the weights projection.
+# Currently the weights projection depends on the sign of scale, so this
+# function does not change the sign (only makes scale 0 in the worst case),
+# which will not cause any issues.
+def finalize_scale_constraints(scale, output_min, output_max):
+  """Clips scale to strictly satisfy all constraints.
+
+  Args:
+    scale: Scale variable of shape: `(units, num_terms)`.
+    output_min: None or minimum layer output.
+    output_max: None or maximum layer output.
+
+  Returns:
+    Clipped scale tensor of same shape as `scale`.
+  """
+  if output_min is not None and output_max is not None:
+    bound = (output_max - output_min) / 2.0
+    scale = tf.clip_by_value(scale, clip_value_min=-bound, clip_value_max=bound)
+  elif output_min is not None:
+    # In this case, we need scale to be nonnegative to properly shift by bias
+    # and satisfy the one-sided max bound.
+    scale = tf.maximum(scale, 0)
+  elif output_max is not None:
+    # In this case, we need scale to be nonpositive to properly mirror and shift
+    # by bias and satisfy the one-sided min bound.
+    scale = tf.minimum(scale, 0)
+  return scale
+
+
 def verify_hyperparameters(lattice_sizes=None,
                            units=None,
                            num_terms=None,
                            input_shape=None,
-                           monotonicities=None):
+                           monotonicities=None,
+                           output_min=None,
+                           output_max=None):
   """Verifies that all given hyperparameters are consistent.
 
   This function does not inspect weights themselves. Only their shape. Use
@@ -337,6 +493,8 @@ def verify_hyperparameters(lattice_sizes=None,
       `monotonicities` is set.
     monotonicities: Monotonicities hyperparameter of `KroneckerFactoredLattice`
       layer. Useful only if `input_shape` is set.
+    output_min: Minimum output of `KroneckerFactoredLattice` layer.
+    output_max: Maximum output of `KroneckerFactoredLattice` layer.
 
   Raises:
     ValueError: If lattice_sizes < 2.
@@ -384,6 +542,12 @@ def verify_hyperparameters(lattice_sizes=None,
                        "'units'. 'units': %s, 'input_shape: %s" %
                        (units, input_shape))
 
+  if output_min is not None and output_max is not None:
+    if output_min >= output_max:
+      raise ValueError("'output_min' must be strictly less than 'output_max'. "
+                       "'output_min': %f, 'output_max': %f" %
+                       (output_min, output_max))
+
 
 def _assert_monotonicity_constraints(weights, units, scale, monotonicities,
                                      eps):
@@ -433,7 +597,110 @@ def _assert_monotonicity_constraints(weights, units, scale, monotonicities,
   return monotonicity_asserts
 
 
-def assert_constraints(weights, units, scale, monotonicities, eps=1e-6):
+def _assert_bound_constraints(weights, units, scale, output_min, output_max,
+                              eps):
+  """Asserts that weights satisfy monotonicity constraints.
+
+  Args:
+    weights: `KroneckerFactoredLattice` weights tensor of shape: `(1,
+      lattice_sizes, units * dims, num_terms)`.
+    units: Number of units per input dimension.
+    scale: Scale variable of shape: `(units, num_terms)`.
+    output_min: None or minimum layer output.
+    output_max: None or maximum layer output.
+    eps: Allowed constraints violation.
+
+  Returns:
+    List of monotonicity assertion ops in graph mode or directly executes
+    assertions in eager mode and returns a list of NoneType elements.
+  """
+  bound_asserts = []
+
+  # Recall that w.shape is (1, lattice_sizes, units * dims, num_terms).
+  weights_shape = weights.get_shape().as_list()
+  _, lattice_sizes, units_times_dims, num_terms = weights_shape
+  assert units_times_dims % units == 0
+  dims = units_times_dims // units
+  weights = tf.reshape(weights, [-1, lattice_sizes, units, dims, num_terms])
+
+  # If both bounds are specified, we must also have that the maximum output be
+  # between -1 and 1.
+  if output_min is not None and output_max is not None:
+    for term, term_weights in enumerate(tf.unstack(weights, axis=4)):
+      max_keypoint_values = tf.reduce_max(
+          tf.abs(term_weights), axis=1, keepdims=True)
+      max_output_values = tf.reduce_prod(
+          max_keypoint_values, axis=3, keepdims=True)
+      for unit, unit_max_output_value in enumerate(
+          tf.unstack(max_output_values, axis=2)):
+        diff = tf.squeeze(1 - unit_max_output_value)
+        bound_asserts.append(
+            tf.Assert(
+                diff >= -eps,
+                data=[
+                    "Bound violation (max output greater than 1)", "Diff", diff,
+                    "Epsilon", eps, "Maximum output value",
+                    unit_max_output_value, "Term index", term, "Unit", unit,
+                    "Weights", weights
+                ]))
+  else:
+    # If only one bound is specified, we must have that all of our weights are
+    # nonnegative at this point. There can be no allowed epsilon error here
+    # because of the effect of a negative value.
+    total_negative_weights = tf.reduce_sum(tf.cast(weights < 0, tf.int32))
+    bound_asserts.append(
+        tf.Assert(
+            total_negative_weights <= 0,
+            data=[
+                "Bound violation (negative weights)",
+                "Number of negative weights", total_negative_weights, "Weights",
+                weights
+            ]))
+
+  # If both bounds are specified, scale must be between
+  # -(output_max-output_min)/2 and (output_max-output_min)/2. If only output_min
+  # is specified, then scale must be nonnegative. If only output_max is
+  # specified, then scale must be nonpositive.
+  if output_min is not None and output_max is not None:
+    bound = (output_max - output_min) / 2.0
+    below_bound_scales = tf.reduce_sum(tf.cast(scale < -bound, tf.int32))
+    above_bound_scale = tf.reduce_sum(tf.cast(scale > bound, tf.int32))
+    bound_asserts.append(
+        tf.Assert(
+            below_bound_scales + above_bound_scale <= 0,
+            data=[
+                "Bound violation (scale out of bounds)", "Bound", bound,
+                "Scale", scale
+            ]))
+  elif output_min is not None:
+    negative_scales = tf.reduce_sum(tf.cast(scale < 0, tf.int32))
+    bound_asserts.append(
+        tf.Assert(
+            negative_scales <= 0,
+            data=[
+                "Bound violation (only output_min specified with negative "
+                "scale values)", "Scale", scale
+            ]))
+  elif output_max is not None:
+    positive_scales = tf.reduce_sum(tf.cast(scale > 0, tf.int32))
+    bound_asserts.append(
+        tf.Assert(
+            positive_scales <= 0,
+            data=[
+                "Bound violation (only output_max specified with positive "
+                "scale values)", "Scale", scale
+            ]))
+
+  return bound_asserts
+
+
+def assert_constraints(weights,
+                       units,
+                       scale,
+                       monotonicities,
+                       output_min,
+                       output_max,
+                       eps=1e-6):
   """Asserts that weights satisfy constraints.
 
   Args:
@@ -442,6 +709,8 @@ def assert_constraints(weights, units, scale, monotonicities, eps=1e-6):
     units: Number of units per input dimension.
     scale: Scale variable of shape: `(units, num_terms)`.
     monotonicities: Monotonicity constraints.
+    output_min: None or minimum layer output.
+    output_max: None or maximum layer output.
     eps: Allowed constraints violation.
 
   Returns:
@@ -459,4 +728,14 @@ def assert_constraints(weights, units, scale, monotonicities, eps=1e-6):
         eps=eps)
     asserts.extend(monotonicity_asserts)
 
+  if output_min is not None or output_max is not None:
+    bound_asserts = _assert_bound_constraints(
+        weights=weights,
+        units=units,
+        scale=scale,
+        output_min=output_min,
+        output_max=output_max,
+        eps=eps)
+    asserts.extend(bound_asserts)
+
   return asserts
